-- file YGREC8/VHDL/ALU8/dec_ALU8.vhdl
-- created dim. janv.  5 03:20:31 CET 2020 by whygee@f-cpu.org (reboot 2020)
--
--
-- Released under the GNU AGPLv3 license
--
-- This unit decodes the instruction word into the followng control signals :
--
--        Func   NEG  PASS_en AND_en XOR_en OrXor_en  CLA_en CMPS  WB CryWr
-- OR     0000     0      0      0      0       1       0     x    1    0
-- XOR    0001     0      0      1      1       1       0     x    1    0
-- AND    0010     0      0      0      1       0       0     x    1    0
-- ANDN   0011     1      0      0      1       0       0     x    1    0
-- CMPU   0100     1      0      1      x       0       1     0    0    1
-- CMPS   0101     1      0      1      x       0       1     1    0    1
-- SUB    0110     1      0      1      x       0       1     0    1    1
-- ADD    0111     0      0      1      x       0       1     0    1    1
-- SET    1000     x      1      1      x       0 pass  0     x    1    0
-- CALL   1001     x      1      1      x       0 pass  0     x    1    0
-- SH/SA  1010     x      0      0      0       0 clear 0     x    1    0
-- RO/RC  1011     x      0      0      0       0 clear 0     x    1   b11 (RC)
-- LDCL/H 1100     x      1      1      x       0 pass  0     x    1    0
-- IN     1101     x      0      0      0       0 clear 0     x    1    0
-- OUT    1110     x      0      0      0       0 clear 0     x    0    0
-- INV    1111     x      0      0      0       0 clear 0     x    0    0
--
-- tentative breakdown into equations :
--
-- CLA_en = F2 & /F3
-- ROP2 = /F3 & /F2
-- F0F1 = F0 | F1
-- /F1F0 = /F1 & F0
-- /F1F3 = F3 & /F1
-- CMPS = CLA_en & /F1F0
-- CarryWrite = CLA_en | opcode=RC
-- RegisterWriteback = /F2 | /F1F3 | (/F3 & F1)
-- OrXor_en = NOR3( F1, F2, F3)
-- XOR_en = F0F1 & ROP2
-- NEG = ( ROP2 & F1 & F0) | (CLA_en & F1F0)
-- Cin = NEG (more or less but also need to decode ADD 0 cond)
-- PASS_en = /F1F3 & /(F2 & F0)
-- AND_en = PASS_en | (/F3 & /F1F0) | CLA_en

Library ieee;
    use ieee.std_logic_1164.all;
Library work;
    use work.all;
Library SLV;
    use SLV.SLV_def.all;
Library YGREC8;
    use YGREC8.ygrec8_def.all;
Library conversions;
    use conversions.conversions.all;

entity DEC_ALU8 is
  port(
    Inst: in SLV16;
    Cin, NEG, Pass_en, AND_en, XOR_en, OrXor_en, CLA_en, CMPS : out SL);
end DEC_ALU8;

architecture LUT of DEC_ALU8 is
  signal add1 : SL;
  alias Opcode: SLV4 is Inst(15 downto 12);
  alias Imm4: SLV4 is Inst(6 downto 3);
  alias FlagIm8: SL is Inst(11);
  alias FlagRI4: SL is Inst(10);

  signal control : SLV7;
  type ControlLUT_type  is array (0 to 15) of SLV7;
  constant ControlLUT : ControlLUT_type := (
     -- NEG  PASS_en AND_en XOR_en OrXor_en  CLA_en CMPS
     "0000100",  -- OR
     "0011100",  -- XOR
     "0001000",  -- AND
     "1001000",  -- ANDN
     "1010010",  -- CMPU
     "1010011",  -- CMPS
     "1010010",  -- SUB
     "0010010",  -- ADD
     "0110000",  -- SET     pass
     "0110000",  -- CALL    pass
     "0000000",  -- SH/SA   clear
     "0000000",  -- RO/RC   clear
     "0110000",  -- LDCL/H  pass
     "0000000",  -- IN      clear
     "0000000",  -- OUT     clear
     "0000000"); -- INV     clear

begin
  add1 <= '1' when Opcode = Op_ADD
        and FlagIm8='0' and FlagRI4='1' and Imm4(3)='0'
     else '0';

  control <= ControlLUT(safe_to_integer(Opcode));
       Cin <= Control(6) or add1;
       NEG <= Control(6);
   Pass_en <= Control(5);
    AND_en <= Control(4);
    XOR_en <= Control(3);
  OrXor_en <= Control(2);
    CLA_en <= Control(1);
      CMPS <= Control(0);

end LUT;

architecture equations of DEC_ALU8 is
  signal add1, CLA, ROP, F0F1, F1F0, F1F3, F2F0, tNEG, Pass : SL;

  alias F3: SL is Inst(15);
  alias F2: SL is Inst(14);
  alias F1: SL is Inst(13);
  alias F0: SL is Inst(12);
  alias Opcode: SLV4 is Inst(15 downto 12);
  alias Imm4: SLV4 is Inst(6 downto 3);
  alias FlagIm8: SL is Inst(11);
  alias FlagRI4: SL is Inst(10);

begin

  -- some partial computations :
   CLA <= F2 and (not F3);
   ROP <= (not F3) and (not F2);
  F0F1 <= F0 or F1;
  F1F0 <= (not F1) and F0;
  F1F3 <= F3 and F1;
  F2F0 <= F2 and F0;

    CLA_en <= CLA;
      CMPS <= CLA and F1F0;
  OrXor_en <= not (F1 or F2 or F3);
    XOR_en <= F0F1 and ROP;
      tNEG <= ( ROP and F1 and F0)
               or (CLA and F1F0);
       NEG <= tNEG;
       Cin <= tNEG or
              ( (not F3) and F2F0 and F1  -- Op_ADD
                  and (not FlagIm8) and FlagRI4 and (not Imm4(3))); -- add positive Imm4
      Pass <= F1F3 and not (F2F0);
   Pass_en <= Pass;
    AND_en <= PASS or ((not F3) and F1F0) or CLA;

end equations;

