-- YGREC8/ASM/Y8asm.vhdl
-- version mer. nov. 10 09:51:56 CET 2021 : the new file assembler
-- version ven. nov. 12 06:05:11 CET 2021 : starts to work
-- version lun. nov. 15 09:45:51 CET 2021 : refactoring
-- version jeu. nov. 18 01:28:48 CET 2021 : removing FWD
-- version jeu. nov. 18 17:47:28 CET 2021 : added DW, 2nd pass OK!
--
-- Released under the GNU AGPLv3 license
--
-- A file-wise 2-pass assembler.
--
--  TODO: sym_PAD, nested expressions, nested contexts...


Library std;
    use std.textio.all;
Library ieee;
    use ieee.std_logic_1164.all;
    use ieee.numeric_std.all;
Library work;
    use work.all;
    use work.ygrec8_asm.all;
Library YGREC8;
    use YGREC8.ygrec8_def.all;
Library dictionary;
    use dictionary.dictionary.all;
Library conversions;
    use conversions.conversions.all;

entity y8asm is
  generic (
    name: string := "";
    dump: string := "";
    verbose: string := ""
  );
end y8asm;


architecture program of y8asm is

  type symbol_type is (
    sym_invalid,
    sym_USR,
    sym_DEF, -- not to be confused with sym_USR
    sym_opcode_alone,
    sym_opcode_Imm9,
    sym_opcode_Imm4,
    sym_opcode_mix,
    sym_reg,
    sym_cond,
    sym_ORG,
    sym_DW,
--    sym_PAD,
--    sym_DB,
--    sym_DS,
--    sym_op1,
--    sym_op2,
--    sym_c_nest,
--    sym_unnest,
    sym_end
  );

  type line_type is (
       line_invalid,  -- just a marker that something shouldn't have happened.
       line_empty,    -- nothing to process on this line.
       line_end,      -- premature end of file
       line_instr,    -- a normal instruction, gets disassembled
       line_label,    --
       line_org,      --
       line_def,      --
       line_dw,       --
       line_miss);    -- a line with an undefined symbol,
                      -- needs to be reassembled during the 2nd pass

  type LineEntry; -- Forward declaration
  type LineEntryPtr is access LineEntry;
  type LineEntry is record
    assembled_value, address, line_number : integer;
    lt : line_type;
    NextEntry : LineEntryPtr;
    txt : line;
  end record;

  shared variable FirstLinePointer : LineEntryPtr;

  shared variable GlobalDict : DictTypePtr;

  function strip_trailspaces(s: string)
    return string is
   variable i: integer :=s'high;
  begin
    while s(i) = ' ' loop
      i := i-1;
      -- no condition to end the loop, fail hard.
    end loop;
    return s(s'low to i);
  end strip_trailspaces;

  procedure generate_symbols is
    variable DEP : DictEntryPtr;

    procedure fix(s: string; t: symbol_type; v: integer := -1) is
    begin
      AddKey(GlobalDict, s, DEP);
      DEP.EntryType  := symbol_type'pos(t);
      DEP.EntryValue := v;
    end fix;

    variable i, j: integer;
    variable str6: string(1 to 6);

  begin
    GlobalDict    := CreateDictionary;

    fix("DEF", sym_DEF);
    fix("ORG", sym_ORG);
    fix("DW" , sym_DW);
    fix("END", sym_end);

    for i in Cond_str_array'range loop
      fix(strip_trailspaces(Cond_str_array(i)), sym_cond, i);
    end loop;

    for i in Reg_str_array'range loop
      fix(Reg_str_array(i), sym_reg, i);
    end loop;

    for i in 0 to 9 loop
      fix(strip_trailspaces(OpcodeS_str_array(i)), sym_opcode_mix, i * 4096);
    end loop;

    for i in 0 to 5 loop
      fix(strip_trailspaces(OpcodeS_str_array(i+10)), sym_opcode_Imm4, (i*2048) + 40960);
    end loop;

    fix("IN" , sym_opcode_Imm9,  13*4096);
    fix("OUT", sym_opcode_Imm9,  14*4096);
    fix("INV", sym_opcode_alone, 65535);
    fix("HLT", sym_opcode_alone, 40959);  -- OPC=Op_CALL, Imm9="111111111", SND=Reg_PC
    fix("NOP", sym_opcode_alone, 0);
    -- Missing : OVL xxh
  end;


  procedure assemble_line(LEP : inout LineEntryPtr;
   current_address : inout integer) is
    -- Input : LEP.txt and LEP.line_number are filled.
    -- current_address may be updated in this procedure,
    -- but don't write LEP.address (it's set on the 1st pass)
    -- Output : LEP.lt at least, LEP.assembled_value if applicable

    variable DEP, DEP2 : DictEntryPtr;
    variable index, lim, len, sri_reg, val, err, opc_add : integer;
    variable c, d : character;
    variable token : string(1 to 16);
    variable symtype : symbol_type;

    procedure skip_spaces(str: string) is
    -- expects index to point at the first character
    -- leave with index updated (or not), and beyond str'high if end reached.
    begin
      parse_space: loop
        if index > str'high then
          return;
        end if;
        c := str(index);
        if c = ';' then  -- start of comment
          index := str'high+1;
          return;
        end if;
        if   (c /= ' ')
         and (c /= HT)
         and (c /= ',')
         and (c /= character'val (160)) then
          return;
        end if;
        index := index + 1;
      end loop parse_space;
    end skip_spaces;

    procedure error_not_EOL(s: string) is
    begin
      index:=lim;
      skip_spaces(LEP.txt.all);
      assert index > LEP.txt'high
        report s & " on line " & integer'image(LEP.line_number)
        severity failure;
    end error_not_EOL;

    procedure error_EOL(s: string) is
    begin
      index:=lim;
      skip_spaces(LEP.txt.all);
      assert index <= LEP.txt'high
        report s & " on line " & integer'image(LEP.line_number)
        severity failure;
    end error_EOL;

    procedure parse_symbol(str: string;
        init: integer:=0) is  -- init allows extra symbols to be parsed at the start of a line.
    -- output: len is the size of the symbol/token, at least 1 (because spaces and EOL must be checked before)
    --    len=0 means an empty token though it should not be possible...
    -- index is expected to point to the first character of the symbol. Not updated.
    -- lim is set to point to the next character after the symbol (either EOL or space)
    -- token(1 to len) is set to the upcased/filtered value/name of the symbol (if no error found)
    begin
      len := token'low-1;
      lim := index;
      parse_sym: loop
        c := str(lim);
        case c is  -- filter some more characters
          when ' ' | HT | ':' |';'| ',' | character'val (160) =>
               exit;
--          when '{' | '}' =>  -- nest, unnest : processed before lookup
          when '$' | '(' | ')' =>  -- here, and expression start and stop
               exit when init /= 0;
               d := c;
          when others => d := upcase_lookup(c);
               assert d/=NUL
                 report "invalid character '" & c & "' on line " & integer'image(LEP.line_number)
                 severity failure;
        end case;
        len := len+1;
        assert len <= token'high
          report "Symbol too long (>16char) on line " & integer'image(LEP.line_number)
          severity failure;
        token(len) := d;
        lim := lim + 1;
        exit when lim > str'high;
      end loop parse_sym;
-- report "token = " & token(1 to len);
    end parse_symbol;

    procedure parse_value(r: inout integer; v: out integer) is
      -- start at "index" (not modified)
      -- modifies lim in some cases
      -- if r=-1 then strict mode is enable (fault on sym not found)
      variable t1, t2 : integer;
    begin
      if LEP.txt(index)='$' then
        r := 0;
        v := current_address;
        lim := index + 1;
        return;
      end if;

      -- TODO : check '(' here

      parse_symbol(LEP.txt.all); -- this modifies lim

      -- check for number
      Num_to_int_p(token(1 to len), t2, t1);
      if t1 = 0 then
        v := t2;
        r := 0;
        return;
      end if;

      FindKey(GlobalDict, token(1 to len), DEP);
      if DEP/=NULL then
        assert DEP.EntryType = symbol_type'pos(sym_USR)
          report "invalid value : reserved word on line " & integer'image(LEP.line_number)
          severity failure;
        v := DEP.EntryValue;
        r := 0;
        return;
      end if;

      assert r /= -1
        report "Symbol " & token(1 to len) &
            " is not defined on line " & integer'image(LEP.line_number)
          severity failure;
      r := -1;
    end parse_value;

    procedure check_new_symbol is
      -- input : token(1 to len)
      -- output : new DEP2
      variable err, i: integer;
    begin
      -- check it's not an existing keyword or symbol:
      FindKey(GlobalDict, token(1 to len), DEP2);
      assert DEP2 = null
        report "Attempt to redefine an existing symbol on line " & integer'image(LEP.line_number)
        severity failure;

      -- check it's not a number either!
      Num_to_int_p(token(1 to len), err, i);
      assert err/=0
        report "Symbol name can't be a number on line " & integer'image(LEP.line_number)
        severity failure;

      for i in 1 to len loop
        assert token(i)  = '_'
           or( token(i) >= '0' and token(i) <= '9' and i > 1)
           or( token(i) >= 'A' and token(i) <= 'Z')
          report "Symbol name contains invalid characters on line " &
                  integer'image(LEP.line_number)
          severity failure;
      end loop;

      AddKey(GlobalDict, token(1 to len), DEP2);
    end check_new_symbol;

    procedure get_dest_register is
    begin
      error_EOL("Missing destination register");
      parse_symbol(LEP.txt.all, 0);
      FindKey(GlobalDict, token(1 to len), DEP);
      assert DEP /= null and DEP.EntryType = symbol_type'pos(sym_reg)
        report "Invalid destination register on line " & integer'image(LEP.line_number)
        severity failure;
      LEP.assembled_value := LEP.assembled_value + DEP.EntryValue;
    end get_dest_register;

    procedure fix_imm(min, max, flag : integer) is
    begin
      assert val >= min and val < max
        report "Immediate value out of range (" & integer'image(val)
             & ") on line " & integer'image(LEP.line_number)
        severity failure;
      if val < 0 then
        val := val + max;
      end if;
      LEP.assembled_value := LEP.assembled_value + (val * 8) + flag;
    end fix_imm;

    procedure end_of_instr is
    begin
      error_not_EOL("spurious text after instruction");
      LEP.lt := line_instr;
      current_address := current_address + 1;
    end end_of_instr;

  begin --assemble_line
    index := LEP.txt'low;
    -- Skip eventual spaces and get the first token:
    skip_spaces(LEP.txt.all);
    if index > LEP.txt'high then
      LEP.lt := line_empty;
      return;
    end if;

    -- get the first word from the line:
    parse_symbol(LEP.txt.all, 1);
    FindKey(GlobalDict, token(1 to len), DEP);

    if DEP=NULL then
      -- test for a label:
      assert lim <= LEP.txt'high and LEP.txt(lim) = ':'
        report "Unknown opcode or command on line " & integer'image(LEP.line_number)
        severity failure;
      lim := lim+1;
      -- check it's a new valid symbol
      check_new_symbol;
      -- and register it like as user defined:
      DEP2.EntryValue := current_address;
      DEP2.EntryType := symbol_type'pos(sym_USR);
      LEP.assembled_value := current_address;
      LEP.lt := line_label;
      error_not_EOL("spurious text after label");
      return;
    end if;

    symtype := symbol_type'val(DEP.EntryType);
    case symtype is

      when sym_end =>
        error_not_EOL("spurious text after END");
        LEP.lt := line_end;
        return;

      when sym_ORG =>
        error_EOL("Missing value for ORG");
        err := -1; -- strict mode
        parse_value(err, val);
        assert err=0
          report "Invalid ORG value on line " & integer'image(LEP.line_number)
          severity failure;
        assert val >= 0 and val < 256
          report "ORG value is out of range (" & integer'image(val)
              & ") on line " & integer'image(LEP.line_number)
          severity failure;
        assert current_address <= val
          report "ORG can't go backwards (line " & integer'image(LEP.line_number) & ")"
          severity failure;
        current_address := val;
        LEP.address := val;
        LEP.lt := line_org;
        error_not_EOL("spurious text after ORG");
        return;

      when sym_DEF =>
        error_EOL("Line ended with no identifier");
        parse_symbol(LEP.txt.all, 0);
        check_new_symbol;
        LEP.lt := line_def;
        error_EOL("Line ended without value");
        err := -1; -- strict mode
        parse_value(err, val);
        assert err=0
          report "value invalid on line " & integer'image(LEP.line_number)
          severity failure;
        LEP.assembled_value := val;
        DEP2.EntryValue := val;
        DEP2.EntryType  := symbol_type'pos(sym_USR);
        error_not_EOL("spurious text after definition");
        return;

      when sym_opcode_alone =>
        error_not_EOL("spurious text after opcode");
        LEP.lt := line_instr;
        LEP.assembled_value := DEP.EntryValue;
        current_address := current_address + 1;
        return;

      when sym_dw =>
        error_EOL("Missing value after DW");
        err := 0;  -- lax mode
        parse_value(err, val);
        if err = -1 then
          LEP.lt := line_miss;
          return;
        end if;

        assert err=0
          report "Invalid immediate value on line " & integer'image(LEP.line_number)
          severity failure;
        assert val >= -32768 and val < 65536
          report "Immediate value out of range (" & integer'image(val)
               & ") on line " & integer'image(LEP.line_number)
          severity failure;
        if val < 0 then
           val := val + 65536;
        end if;
        LEP.assembled_value := val;
        error_not_EOL("spurious text after value");
        LEP.lt := line_dw;
        current_address := current_address + 1;
        return;

      when sym_opcode_Imm9 =>
        error_EOL("Missing value after opcode");
        LEP.assembled_value := DEP.EntryValue;

        err := 0;  -- lax mode
        parse_value(err, val);
        if err = -1 then
          LEP.lt := line_miss;
          return;
        end if;

        assert err=0
          report "Invalid immediate value on line " & integer'image(LEP.line_number)
          severity failure;
        assert val >= -256 and val < 512
          report "Immediate value out of range (" & integer'image(val)
               & ") on line " & integer'image(LEP.line_number)
          severity failure;
        if val < 0 then
           val := val + 512;
        end if;
        LEP.assembled_value := LEP.assembled_value + (val * 8);
        get_dest_register;
        end_of_instr;
        return;

      when sym_opcode_Imm4 | sym_opcode_mix =>
        error_EOL("Missing value after opcode");
        LEP.assembled_value := DEP.EntryValue;

        opc_add := 0;
        if DEP.EntryValue = 28672 then -- ADD ? (for eventual off-by-one later)
          opc_add := 1;
        end if;

        -- get the 2nd argument and look it up against the register names.
        sri_reg := 0;
        parse_symbol(LEP.txt.all, 0);
        FindKey(GlobalDict, token(1 to len), DEP);
        if DEP /= null and DEP.EntryType = symbol_type'pos(sym_reg) then
          sri_reg:=1;
          LEP.assembled_value := LEP.assembled_value + (DEP.EntryValue * 8);
        else
          parse_value(err, val);
          if err = -1 then
            LEP.lt := line_miss; -- try again next time...
            return;
          end if;
          assert err = 0
            report "Invalid immediate value on line " & integer'image(LEP.line_number)
            severity failure;
              -- the value will be processed in the condition section
        end if;

        get_dest_register;

        -- is there a condition ?
        index:=lim;
        skip_spaces(LEP.txt.all);
        if index > LEP.txt'high then
          -- No condition => post-process eventual IMM(4/8):
          if sri_reg=0 then
            if symtype = sym_opcode_Imm4 then
              fix_imm(-8, 16, 1024);
              LEP.assembled_value := LEP.assembled_value + 512;  -- set the N bit.
            else
              fix_imm(-128, 256, 2048); -- no condition: we can have a whole byte as immediate value
            end if;
          else
            -- no imm, so it's a reg source, so set the condition field to ALWS
            LEP.assembled_value := LEP.assembled_value + 512;  -- set the N bit.
          end if;
        else
          -- expect a condition keyword
          parse_symbol(LEP.txt.all, 0);
          FindKey(GlobalDict, token(1 to len), DEP);
          assert DEP /= null and DEP.EntryType = symbol_type'pos(sym_cond)
            report "Expected a condition on line " & integer'image(LEP.line_number)
            severity failure;

          LEP.assembled_value := LEP.assembled_value + (DEP.EntryValue*64);

          if sri_reg=0 then
            assert (DEP.EntryValue mod 2) = 0
              report "Invalid condition for Imm4 on line " & integer'image(LEP.line_number)
              severity failure;

            -- fix a corner case
            if opc_add = 1 then  
              assert val /= 0
                report "Short ADD 0 is not possible ! on line " & integer'image(LEP.line_number)
                severity failure;
              if val > 0 then
                val := val - 1;
              end if;
            end if;
            fix_imm(-8, 16, 1024);
          end if;
          error_not_EOL("spurious text after condition");
        end if;

        LEP.lt := line_instr ;
        current_address := current_address + 1;
        return;

      when others =>
        report token(1 to len) & " is not yet supported (line " & integer'image(LEP.line_number) & ")"
                severity failure;
    end case;
  end assemble_line;

  procedure import_asm is
    file f : text open read_mode is name;
    variable LEP, last_line : LineEntryPtr;
    variable current_address, line_count : integer := 0;
  begin
    FirstLinePointer := null;
    LEP := new LineEntry;
    while (not endfile(f)) loop
      line_count := line_count + 1;
      readline(f, LEP.txt);
      LEP.line_number := line_count;
      LEP.lt := line_invalid;
      LEP.address := current_address;
      assemble_line(LEP, current_address);  -- current_address might be updated.
      assert LEP.lt /= line_invalid
        report "ooopsie, internal error in first pass (LEP.lt = line_invalid) on line " & integer'image(LEP.line_number)
        severity failure;

      if LEP.lt /= line_empty then
        -- register the line:
        if FirstLinePointer = null then
           FirstLinePointer := LEP;
        else
          last_line.NextEntry := LEP;
        end if;
        last_line := LEP;
        exit when LEP.lt = line_end;
        assert current_address < 256
          report "instruction address > 255 on line " & integer'image(LEP.line_number)
        severity failure;
        LEP := new LineEntry;
      end if;
    end loop;
  end;

  procedure export_hyx is
    file xf : text open write_mode is name & ".hyx";
    variable LEP : LineEntryPtr;
    variable l, v : integer;
    variable inst : std_logic_vector(15 downto 0);
    variable result: string(1 to 18);
    variable ALTP : AlphaLookupTypePtr;
    variable DEP : DictEntryPtr;

  begin
    -- output the signature :
    write (xf, ";;hyx1" & LF);

    -- scan all the stored lines
    LEP := FirstLinePointer;
    while LEP /= null loop
      assert LEP.lt /= line_invalid
        report "ooopsie, internal error (LEP.lt = line_invalid) on line " & integer'image(LEP.line_number)
        severity failure;  -- hey, you never know.
      assert LEP.lt /= line_empty
        report "ooopsie, internal error (LEP.lt = line_empty) on line " & integer'image(LEP.line_number)
        severity failure;  -- hey, you never know.

      if LEP.lt = line_miss then
        -- try to reassemble the line again here,
        -- hoping that the symbol is resolved this time.
        l := LEP.address;  -- handle the case of $
        assemble_line(LEP, l);
        -- l is not considered on the 2nd pass, it's a dummy.
        assert LEP.lt /= line_miss
          report "unresolved symbol on line " & integer'image(LEP.line_number)
          severity failure;
      end if;

      write (xf, "; L" & integer'image(LEP.line_number) & ": " & LEP.txt.all & LF);
      case LEP.lt is
        when line_org =>
          write (xf, ':' & int2hexstr(LEP.address, 15, 2) & LF);

        when line_def | line_label =>
          write (xf, "; = " & integer'image(LEP.assembled_value) & LF);

        when line_dw =>
          write (xf, int2hexstr(LEP.assembled_value, 4, 4) & " ; @" & integer'image(LEP.address) & LF);

        when line_instr =>
          v := LEP.assembled_value;
          assert v >= 0 and v < 65536   -- hey you never know.
            report "ooopsie, instruction out of range on line " & integer'image(LEP.line_number)
            severity failure;
          inst := std_logic_vector(to_unsigned(v, 16));
          Disassemble(inst, "", result);
          -- write 16 bits
          write (xf, int2hexstr(v, 4, 4) & " ; @" & integer'image(LEP.address) & ": " & result & LF);

        when others => null;
      end case;

      LEP := LEP.NextEntry;
    end loop;

    -- Dump the symbol table:
    if dump/="" then
      write (xf, LF & ";;;; SYMBOL DUMP :" & LF);
      for l in DictType'range loop
        ALTP := GlobalDict(l);
        if ALTP /= NULL then
          for c in AlphaLookupType'range loop
            DEP := ALTP(c);
            while DEP /= NULL loop
              if (DEP.EntryType = symbol_type'pos(sym_USR))
               or dump = "full" then
                write (xf, "; * '" & string(DEP.Key(1 to l))
                  & "'=" & integer'image(DEP.EntryValue)
                  & " ref:" & integer'image(DEP.RefCount)
                  & " / " & symbol_type'image(symbol_type'val(DEP.EntryType))
                  & LF);
              end if;
              DEP := DEP.NextEntry;
            end loop;
          end loop;
        end if;
      end loop;
    end if;
  end;

begin
  main: process
  begin
    assert name /= ""
      report LF & "please specify a source file name with"
         & LF & "the command line parameter -gname=sourcefile.y8" & LF
      severity failure;
    generate_symbols;
    import_asm;  -- 1st pass
    export_hyx;  -- 2nd pass
    wait;
  end process;
end program;
