-- YGREC8/ASM/ygrec8_asm.vhdl
-- version mer. déc. 26 08:39:57 CET 2018 : reboot
-- version sam. déc. 29 23:46:13 CET 2018
-- version jeu. mars 14 15:51:29 CET 2019 : new format
-- Released under the GNU AGPLv3 license
--
-- A collection of functions to assemble and
-- disassemble binary code for the YGREC8

Library ieee;
    use ieee.std_logic_1164.all;
    use ieee.numeric_std.all;
Library conversions;
    use conversions.conversions.all;
Library SLV;
    use SLV.SLV_def.all;
Library YGREC8;
    use YGREC8.ygrec8_def.all;

package ygrec8_asm is
  procedure Disassemble(inst: SLV16; context: string; result: inout string(1 to 18));
  function Assemble(txt, context: string) return SLV16;

  -- used both for validating alphanumerical characters
  -- and providing an uppercase function
  type char_array_type is array (character) of character;
  constant upcase_lookup : char_array_type := (
   '0'=>'0',  '1'=>'1',  '2'=>'2',  '3'=>'3',  '4'=>'4',
   '5'=>'5',  '6'=>'6',  '7'=>'7',  '8'=>'8',  '9'=>'9',
   'a'=>'A',  'b'=>'B',  'c'=>'C',  'd'=>'D',  'e'=>'E',  'f'=>'F',  'g'=>'G',
   'h'=>'H',  'i'=>'I',  'j'=>'J',  'k'=>'K',  'l'=>'L',  'm'=>'M',  'n'=>'N',
   'o'=>'O',  'p'=>'P',  'q'=>'Q',  'r'=>'R',  's'=>'S',  't'=>'T',  'u'=>'U',
   'v'=>'V',  'w'=>'W',  'x'=>'X',  'y'=>'Y',  'z'=>'Z',
   'A'=>'A',  'B'=>'B',  'C'=>'C',  'D'=>'D',  'E'=>'E',  'F'=>'F',  'G'=>'G',
   'H'=>'H',  'I'=>'I',  'J'=>'J',  'K'=>'K',  'L'=>'L',  'M'=>'M',  'N'=>'N',
   'O'=>'O',  'P'=>'P',  'Q'=>'Q',  'R'=>'R',  'S'=>'S',  'T'=>'T',  'U'=>'U',
   'V'=>'V',  'W'=>'W',  'X'=>'X',  'Y'=>'Y',  'Z'=>'Z',
   '-'=>'-',  '_'=>'_',   others=>NUL);

end ygrec8_asm;

package body ygrec8_asm is

  procedure Disassemble(inst: SLV16; context: string; result: inout string(1 to 18)) is
    variable srii, sndi: integer;

    alias OPC  : SLV4 is inst(15 downto 12);
    alias Imm9 : SLV9 is inst(11 downto  3);
    alias Imm8 : SLV8 is inst(10 downto  3);
    alias Imm4 : SLV4 is inst( 6 downto  3);
    alias CND4 : SLV4 is inst( 9 downto  6);
    alias CND3 : SLV3 is inst( 8 downto  6);
    alias SRI  : SLV3 is inst( 5 downto  3);
    alias SND  : SLV3 is inst( 2 downto  0);

    variable CND : SLV4 := CND4;
  begin
    result:=(others=>' ');

    -- start with aliases and the most special cases
    if inst="0000000000000000" then
      result(1 to 3):="NOP";
      return;
    end if;

    if OPC=Op_CALL and SND=Reg_PC then
      if Imm9="111111111" then
        result(1 to 3):="HLT";
      else
        result(1 to 7):="OVL " & SLV_to_Hex(Imm8) & "h";
-- /!\ bit 11 not used ?
      end if;
      return;
    end if;

    result(1 to 6) := Disasm_str_array(safe_to_integer(inst(15 downto 11)));
    -- result(6) contains a "flag" for later.

    if OPC=Op_INV then
      return;  -- no argument at all
    end if;

    sndi := safe_to_integer(SND); -- precomputed for later

    if OPC=Op_IN     -- the only opcodes with Imm9
    or OPC=Op_OUT then
      result(5 to 11) := SLV_to_Hex(Imm9)
                       & "h "
                       & Reg_str_array(sndi);
      return;
    end if;

    -- Check Imm8
    if result(6)=' ' and inst(11)='1' then
      result(6 to 11) := SLV_to_Hex(Imm8)
                       & "h "
                       & Reg_str_array(sndi);
      return;
    end if;

    if inst(10)='1' then  --  SRI/Imm4 ?

      -- Imm4
      CND(0) := '0'; -- disable extended conditions
      srii := safe_to_integer(Imm4);
      if inst(6)='1' then  -- negative ?
        srii := 16-srii; -- fold for sign extension, and fit in a single character
        result(6) := '-';
      else
        if OPC=Op_ADD then
          srii := srii+1;  -- done by forcing the carry bit
        end if;
        result(6) := ' ';
      end if;
      result(7 to 7):= integer'image(srii);  -- no overflow expected here, by design

    else
      -- Reg
      result(6 to 7):=Reg_str_array(safe_to_integer(SRI));
    end if;

    result(9 to 10):= Reg_str_array(sndi);

    if CND/=Cond_A then
      result(12 to 15):= Cond_str_array(safe_to_integer(CND));
    end if;

    return;
  end Disassemble;

---------------------------------------------------------------------------------------

  function Assemble(txt, context: string) return SLV16 is
    variable instruction_slv: SLV16 := "UUUUUUUUUUUUUUUU"; -- invalid result
    variable token : string (1 to 12);
    variable err, i, index, strlen, has_Imm8, has_SRI, has_SND, has_cond,
              SRI_is_Imm4, SRI_is_reg, is_ADD : integer := 0;
    variable CND4 : SLV4;

    -- populate token with an upcased word
    -- leave with strlen=0 if end of line, -1 if error
    procedure parse_token(strlen: inout integer) is
      variable c, d : character;
    begin
      token := "            ";
      strlen:=0;

      -- skip the leading spaces
      parse_space: loop
        if index > txt'high then return; end if;
        c := txt(index);
        index := index + 1;
        exit when (c /= ' ') and (c /= HT);
      end loop parse_space;

      -- terminate on comment:
      if c=';' then return; end if;

      parse_word: loop
        d := upcase_lookup(c);
        if d=NUL then
          report context & "|Parse : invalid character '"
             & c & "' (" & integer'image(character'pos(c))
             & ") at index " & integer'image(index-1);
          strlen := -1;
          return;
        end if;

        strlen := strlen +1;
        if strlen > 12 then
          report context & "|Parse : " & token --(1 to 12)
             & "... : word is longer than 12 characters";
          strlen := -1;
          return;
        end if;
        token(strlen) := d; -- finally copy the upcase character

        if index > txt'high then return; end if;
        c := txt(index);
        index := index + 1;
        if (c = ' ') or (c = HT) or (c = ';')
          then return; end if;
      end loop parse_word;
    end parse_token;

    procedure parse_Reg(name: string(1 to 2); offset: integer; result: out integer) is
      variable i: integer :=Reg_str_array'high;
    begin
      loop
        if Reg_str_array(i)=name then
          instruction_slv(offset+2 downto offset) := std_logic_vector(to_unsigned(i,3));
          result:=i;
          return;
        end if;

        i := i-1;
        exit when i < 0;
      end loop;

      result:=-1;
      return;
    end parse_Reg;

  begin

    check_more_text: loop   -- not real loops, just a structure
      check_source: loop    -- so I can use exit to replace goto
        check_opcode: loop

          -- get the first word: must be the OPCODE
          index := txt'low;
          parse_token(strlen);
          if strlen < 1 then
            return "UUUUUUUUUUUUUUUU";
          end if;

          -- short opcode aliases :
          if strlen=3 then
            if token(1 to 3)="NOP" then
              instruction_slv := "0000000000000000";
              exit check_more_text;
            end if;
            if token(1 to 3)="INV" then
              instruction_slv := "1111111111111111";
              exit check_more_text;
            end if;
            if token(1 to 3)="HLT" then
              instruction_slv := Op_CALL & "111111111111";
              exit check_more_text;
            end if;

            -- longer instructions with more arguments :
            if token(1 to 3)="OVL" then
              has_Imm8 := 1;
              instruction_slv(15 downto 12) := Op_CALL;
              instruction_slv(2 downto 0) := Reg_PC;
              exit check_opcode;
            end if;
          end if;

          if (strlen=2 and token(1 to 2)="IN")
          or (strlen=3 and token(1 to 3)="OUT") then
            has_SND := 1; 
            if strlen=3 then
              instruction_slv(15 downto 12) := Op_OUT;
            else
              instruction_slv(15 downto 12) := Op_IN;
            end if;
-- Imm9
            parse_token(strlen);   -- could have been merged with Imm8 ?
            Num_to_int_p(token(1 to strlen), i, err);
            if err /= 0 then
              report context & "|Parse : missing or wrong argument for Imm9: '" & token(1 to strlen) & "'";
              return "UUUUUUUUUUUUUUUU";
            end if;
            if (i < -256) or (i > 511) then
              report context & "|Parse : Imm9 out of range: '" & token(1 to strlen) & "'";
              return "UUUUUUUUUUUUUUUU";
            end if;
            if (i<0) then
              i := i+512; -- wrap around
            end if;
            instruction_slv(11 downto 3) := std_logic_vector(to_unsigned(i,9));
            exit check_source; -- get SND
          end if;

          -- the opcode is not one of the special cases so look it up in the table
          if (strlen>=2) and (strlen<=4) then -- constrain the size of the opcode word
            i:=15; -- OpcodeS_str_array'high;  -- skip IN, OUT and INV
-- opcode
            scan_opcodes: loop
              if token(1 to 4) = OpcodeS_str_array(i) then
                instruction_slv(15 downto 11) := OpcodeS_val_array(i);
                if instruction_slv(11) = 'X' then
                  instruction_slv(11) := '0'; -- default to RI4 form,
                   -- Imm8 must change it if found, but not RI4 (which would overwrite the LSB of the opcode )
                  has_Imm8 := 1;
                end if;
                has_SRI := 1;
                has_SND := 1;
                if token(1 to 4) = "ADD " then
                  is_ADD := 1; -- for later, add an offset to imm4
                end if;
                exit check_opcode;
              end if;

              i := i-1;
              exit when i < Opcode_str_array'low;
            end loop scan_opcodes;
          end if;
          report context & "|Parse : unknown opcode '" & token(1 to strlen) & "'";
          return "UUUUUUUUUUUUUUUU";

        end loop check_opcode;

        parse_token(strlen); -- get the argument
-- SRI
        -- First, try to find a register
        if has_SRI=1 and strlen=2 then
          parse_Reg(token(1 to 2), 3, i);
          if i >= 0 then
            instruction_slv(10 downto 6) := '0' & Cond_A; -- default=Always, register form
            SRI_is_reg := 1;
            exit check_source;
          end if;
        end if;

        if (has_SRI=1 or has_Imm8=1) and strlen>0 then
          -- if it's not a register, it can only be a number
          Num_to_int_p(token(1 to strlen), i, err);
          if err /= 0 then
            report context & "|Parse : missing or wrong number at '" & token(1 to strlen) & "'";
            return "UUUUUUUUUUUUUUUU"; -- return instruction_slv;
          end if;

-- Imm4
          if (has_SRI=1) then
            try_imm4: loop
              if (i>= -8) and (i <= 7+is_ADD) then
                if i < 0 then
                  i := i+16;  -- wraparound
                else 
                  if is_ADD=1 then
                    exit try_imm4 when i=0; -- try as imm8 because this code is removed
                    i := i-1;
                  end if;
                end if;
                instruction_slv(10 downto 3) := '1' & Cond_A(3 downto 1) & std_logic_vector(to_unsigned(i,4));
                SRI_is_Imm4 := 1;
                exit check_source;
              else
                if has_Imm8=0 then
                  report context & "|Parse : Imm4 out of range '" & token(1 to strlen) & "'";
                  return "UUUUUUUUUUUUUUUU"; -- return instruction_slv;
                -- else : try our luck below
                end if;
              end if;
              exit try_imm4;
            end loop try_imm4;
          end if;

-- Imm8
          assert has_Imm8=1 severity failure;  -- here it can only be Imm8
          if (i > 255) or (i < -128) then 
            report context & "|Parse : Imm8 out of range '" & token(1 to strlen) & "'";
            return "UUUUUUUUUUUUUUUU"; -- return instruction_slv;
          end if;
          if i < 0 then
            i := i+256;  -- wraparound
          end if;
          instruction_slv(11 downto 3) := '1' & std_logic_vector(to_unsigned(i,8));
          exit check_source;
        end if;

        -- complain because neither REG or IMMM
        report context & "|Parse : missing or wrong source argument '" & token(1 to strlen) & "'";
        return "UUUUUUUUUUUUUUUU"; -- return instruction_slv;
      end loop check_source;

-- SND
      if has_SND=1 then
        parse_token(strlen);
        lookup_SND: loop
          if strlen=2 then
            parse_Reg(token(1 to 2), 0, i);
            exit lookup_SND when i >= 0;
          end if;
          report context & "|Parse : missing or wrong SND register argument '" & token(1 to strlen) & "'";
          return "UUUUUUUUUUUUUUUU"; -- return instruction_slv;
        end loop lookup_SND;
      end if;

-- Condition (optional, trickier)
      if SRI_is_Imm4=1 or SRI_is_reg=1 then
        parse_token(strlen);
        if strlen = -1 then return "UUUUUUUUUUUUUUUU"; end if;
        if strlen =  0 then return instruction_slv;    end if; -- nothing left to check;
        if (strlen >= 3) and (strlen <= 4) then
          i := 0;
          check_cond: loop
            if token(1 to 4) = Cond_str_array(i) then
              CND4 := std_logic_vector(to_unsigned(i,4));
              instruction_slv(9 downto 6+SRI_is_Imm4) := CND4(3 downto SRI_is_Imm4);
              exit check_more_text;
            end if;
            i := i+1+SRI_is_Imm4; -- skip unavailable conditions
            exit when i > Cond_str_array'high;
          end loop check_cond;
        end if;
        -- complain that nothing matches
        report context & "|Parse : invalid condition argument '" & token(1 to strlen) & "'";
        return "UUUUUUUUUUUUUUUU"; -- return instruction_slv;
      end if;
      exit;

    end loop check_more_text;

    -- check there is no more text:
    parse_token(strlen);
    if strlen = -1 then return "UUUUUUUUUUUUUUUU"; end if;
    if strlen > 0 then
      report context & "|Parse : spurious characters after instruction : '" & token(1 to strlen) & "'";
      return "UUUUUUUUUUUUUUUU";
    end if;

    return instruction_slv;
  end Assemble;

end ygrec8_asm;
